/*
** Module   : p2j.screen.js
** Abstract : GUI-specific screen management
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20150325 Created GUI version from ChUI version. Right now it is not functional.
**                  Renamed objects to match the more generic web client approach (not just
**                  ChUI). Moved some logic to common code locations.
** 002 SBI 20150731 Applied "use strict" directive, fixed undeclared variables.
** 003 GES 20150720 Total rewrite to implement a first pass at GUI support,
**                  added a drawing image operation.
** 004 GES 20150818 Added more GUI support including fonts, cursors and z-order.
*/

"use strict";

p2j.screen = (function()
{
   /** Screen object which will be returned from this initialization function. */
   var me = {};
   
   /** Color scaling factor, compatible with Java2D. */
   var SCALE_FACTOR = 0.7;
   
   /** Minimum color value for darkening/lightening. */
   var MIN_COLOR_VALUE = Math.floor(1.0 / (1.0 - SCALE_FACTOR));
   
   /** List of active windows. */
   var winlist = new Array();

   /** Windows in z-order (lowest at index 0 to highest at length - 1). */
   var zlist = new Array();

   /** Loaded images map */
   var loadedImages = new Map();
   
   /** The current font ID, in the js font-table. */
   var currentFont;
   
   /** PaintPrimitives to ordinal mappings. Change this to match Java PaintPrimitives changes! */
   var ops = 
   {
      DRAW_STRING          : 0,
      DRAW_STRING_SCALED   : 1,
      DRAW_PARAGRAPH       : 2,   
      DRAW_RECT            : 3,
      DRAW_ROUND_RECT      : 4,
      DRAW_IMAGE           : 5,
      FILL_RECT            : 6,
      FILL_ROUND_RECT      : 7,
      FILL_POLYGON         : 8,
      DRAW_LINE            : 9,
      DRAW_3D_RECT         : 10,
      FILL_3D_RECT         : 11,
      SET_COLOR            : 12,
      SET_LINE_STROKE      : 13,
      SET_FONT             : 14,
      SET_FONT_STYLE       : 15,
      TRANSLATE_PUSH       : 16,
      TRANSLATE_POP        : 17,
      CLEAR                : 18, 
      CLEAR_AREA           : 19,
      NO_CLIP              : 20,
      CLIP                 : 21,
      FLUSH                : 22,
      SET_LOCATION         : 23,
      RESIZE_WINDOW        : 24,
      DRAW_HIGHLIGHT       : 25,
      CLEAR_HIGHLIGHT      : 26,
      START_BATCH          : 27,
      END_BATCH            : 28,
      SET_TITLE            : 29
   };
   
   // reverse map the types to their property names
   for (var oi in ops)
   {
      if (ops.hasOwnProperty(oi))
      {
         ops[ops[oi]] = oi;
      }
   }

   /**
    * Creates a line strokes object that plays a manager role, used to apply stroke styles
    * to lines and to the native canvas context.
    */
   var strokesManager = new LineStrokes();
   
   /**
    * Constructor for a Window class that represents a Progress 4GL window instance.
    *
    * @param   wid
    *          The window id by which this window can be uniquely identified.
    * @param   settings
    *          Optional settings for the window.
    */
   function Window(wid, settings)
   {
      /** Window id by which this window can be uniquely identified. */
      this.id = wid;
      
      var options = settings || { };
      
      options.id     = options.id    || ("pwindow_" + wid);
      options.width  = options.width || 400;
      options.height = options.width || 600;
      options.style  = options.style || { };
      
      // this must always be set so that the canvas element is removed from the normal layout
      // processing of the document
      options.style.position = "absolute";
      
      /** Backing canvas on which all output for this window will be rendered. */
      this.canvas = p2j.createCanvas(null, options);
      
      /** Canvas 2D graphics context. */
      this.ctx = this.canvas.getContext('2d');
      
      /** List of owned windows. */
      this.children = new Array();
      
      /** Stack of graphics context state, used for managing drawing, clipping and transforms. */
      this.state = new Array();
      
      /** Current color for direct drawing (in the form [ red, green, blue ]). */
      this.rawColor = new Array(0, 0, 0);
      
      /** Common data structure used in direct drawing operations. */
      this.pixel = this.ctx.createImageData(1, 1);
      
      /** Internal portion of the pixel data structure used in direct drawing operations. */
      this.pixData = this.pixel.data;
      
      // force alpha to 100% opaque
      this.pixData[3] = 0xFF;
      
      // force the line width default
      this.ctx.lineWidth = 1;
      
      // eliminate anti-aliasing of images
      this.ctx.imageSmoothingEnabled = false;
      
      // the current window stroke style
      this.strokeStyle = strokesManager.getDefaultStrokeStyle();

      // the current window stroke width
      this.strokeWidth = 1;

      this.origin = { "x" : 0, "y" : 0 };
      
      // force all drawing to be inside pixel "cells" instead of "stradling" between two cells
      // which is the default; this eliminates much of the worst of the anti-aliasing and
      // positioning problems inherent in canvas operations; this will also save off the
      // current state of the graphics context so that it can be restored later 
      this.translate(0.5, 0.5);
   };
   
   /**
    * Cleanup any resources, including any linkages to the window from an owning window.
    */
   Window.prototype.cleanup = function()
   {
      // remove the child linkage from the owner if it exists
      if (isExistingWindowId(this.owner))
      {
         var parent = getWindow(this.owner);
         
         for (var prop in parent.children)
         {
            if (parent.children[prop] === this.id)
            {
               delete parent.children[prop];
               break;
            }
         }
      }
      
      // remove from the DOM
      this.parentNode.removeChild(this);
      
      // TODO: compress the zIndex values and manage z-order state
   };
   
   /**
    * Translate the graphics context by the given X and Y factors.  This will also save the state
    * of the context before the translate is applied, so that the translate values can be
    * reversed later.
    *
    * @param    {Number} x
    *           X dimension factor.
    * @param    {Number} y
    *           Y dimension factor.
    */
   Window.prototype.translate = function(x, y)
   {
      this.save({ translate : { "x" : x, "y" : y } });
      
      this.origin.x += x;
      this.origin.y += y;
      
      this.ctx.translate(x, y);
   };
   
   /**
    * Untranslate the graphics context by the given X and Y factors. This does a restore on
    * the previously saved context state.  It will error out if some sanity checks don't pass.
    * Make sure that calls to this function are always properly paired with equal and opposite
    * calls to translate().
    *
    * @param    {Number} x
    *           X dimension factor.
    * @param    {Number} y
    *           Y dimension factor.
    */
   Window.prototype.untranslate = function(x, y)
   {
      var top = this.peek();
      
      if (top.translate)
      {
         if (top.translate.x === -x && top.translate.y === -y)
         {
            // clear our state back to the version that doesn't have this translate
            this.restore();
            
            this.origin.x += x;
            this.origin.y += y;
         }
         else
         {
            me.error("Mismatching translate pop values, expected (" +
                     top.translate.x.toString() + ", " + top.translate.y.toString() +
                     ") and found " + x.toString() + ", " + y.toString() + ").");
         }
      }
      else
      {
         me.error("Out of order translate pop!");
      }
   };
   
   /**
    * Add the given rectangle to the current clipping region.
    * 
    * @param    {Number} x
    *           The top left X coordinate in pixels.
    * @param    {Number} y
    *           The top left Y coordinate in pixels.
    * @param    {Number} width
    *           The rectangle width in pixels.
    * @param    {Number} height
    *           The rectangle height in pixels.
    */
   Window.prototype.clip = function(x, y, width, height)
   {
      this.save({ clip : { "x" : x, "y" : y, "width" : width, "height" : height } });
      this.ctx.beginPath();
      this.ctx.rect(x, y, width, height);
      this.ctx.clip();
   };
   
   /**
    * Remove the most recently added rectangle from the current clipping region. The graphics
    * context state is popped to restore the previously used values.
    */
   Window.prototype.unclip = function()
   {
      this.restore();
   };
   
   /**
    * Save graphics context state to the stack.
    *
    * @param    {Object} additional
    *           Caller provided name/value pairs to add to the state.
    */
   Window.prototype.save = function(additional)
   {
      var cur = 
      {
         lineWidth   : this.ctx.lineWidth,
         fillstyle   : this.ctx.fillStyle,
         strokeStyle : this.ctx.strokeStyle
      };
      
      if (additional)
      {
         for (var prop in additional)
         {
            if (additional.hasOwnProperty(prop))
            {
               cur[prop] = additional[prop];
            }
         }
      }
      
      this.state.push(cur);
      
      // now force the context to push its state internally (only the context itself can save off
      // its accumulated clip region and its cumulative transforms)
      this.ctx.save();
   };
   
   /**
    * Restore the graphics context state to the state as stored on the top of the stack.
    * <p>
    * The state in our own custom state stack is not actually used for the restore, instead the
    * context's internal stack is used (ctx.restore()).  Our stack is used for other purposes
    * and we simply clear it here.
    */
   Window.prototype.restore = function()
   {
      this.state.pop();
      this.ctx.restore();
   };
   
   /**
    * Obtain the graphics context state as stored on the top of the stack.
    *
    * @return   The top of state stack object.
    */
   Window.prototype.peek = function()
   {
      var last = this.state.length - 1;
      
      return last >= 0 ? this.state[last] : {};
   };
   
   /**
    * Replay the saved graphics context state back into a newly reset graphics context. This
    * should only be used after a canvas resize operation.
    */
   Window.prototype.replay = function()
   {
      var len = this.state.length;
      
      for (var i = 0; i < len; i++)
      {
         var next = this.state[i];
         
         this.ctx.lineWidth   = next.lineWidth;
         this.ctx.fillstyle   = next.fillStyle;
         this.ctx.strokeStyle = next.strokeStyle;
         
         var x;
         var y;
         var width;
         var height;
         
         if (next.translate)
         {
            x = next.translate.x;
            y = next.translate.y;
            
            // don't call the this.translate() method here because it modifies the stack that
            // we are simply trying to replay; we don't want the stack to be modified
            this.ctx.translate(x, y);
         }
      
         if (next.clip)
         {
            x      = next.clip.x;
            y      = next.clip.y;
            width  = next.clip.width;
            height = next.clip.height;                               
                                                                     
            // don't call the this.translate() method here because it modifies the stack that
            // we are simply trying to replay; we don't want the stack to be modified
            this.ctx.beginPath();
            this.ctx.rect(x, y, width, height);
            this.ctx.clip();
         }
      }
   };
   
   /**
    * Set the color for future graphics operations on this window.
    *
    * @param    {Number[]} or {String} color
    *           This must be either a valid array of 3 integers between 0 and 255 inclusive OR it
    *           an HTML color specification in "#RRGGBB" form.
    */
   Window.prototype.setColor = function(color)
   {
      var colorSpec;
   
      if (color instanceof Array)
      {
         if (validateColor(color))
         {
            colorSpec = createColorString(color);
            
            // store the color in a form that can be used for direct drawing
            this.rawColor = [ color[0], color[1], color[2] ];
         }
         else
         {
            me.error("Invalid array form of color specification.");
            return;
         }
      }
      else
      {
         if (typeof color === "string")                  
         {
            colorSpec = color;
            
            // store the color in a form that can be used for direct drawing
            this.rawColor = parseColor(color);
         }
         else
         {
            me.error("Invalid string form of color specification.");
            return;
         }
      }
      
      // change the canvas colors
      this.ctx.fillStyle   = colorSpec;
      this.ctx.strokeStyle = colorSpec;
   };
   
   /**
    * Show or hide the window.
    *                              
    * @param    {boolean} visible  
    *           Flag that determines the new visibility state.
    */
   Window.prototype.setVisible = function(visible)
   {
      // this code depends on each canvas being style.position = 'absolute' so that it is
      // outside of the browser's layout processing
      this.canvas.style.visibility = visible ? "visible" : "hidden";
      
      // bring to the top of the z-order
      moveZOrderEntryToTop(this.id);
   };
   
   /**
    * Resize the window to the given width and height.
    *
    * @param    {Number} width
    *           The new window width in pixels.
    * @param    {Number} height
    *           The new window height in pixels.
    */
   Window.prototype.resize = function(width, height)
   {
      // save off the previously drawn content
      var oldPixels = this.ctx.getImageData(0, 0, this.canvas.width, this.canvas.height);
      
      // this is expensive, but we really do need to change the overall size of the canvas 
      this.canvas.width  = width;
      this.canvas.height = height;
      
      // all of the current state including colors, line stroke, translation and clipping
      // must be reapplied to the context because changing the size resets everything
      this.replay();
                                   
      // re-blit the previously drawn content
      this.ctx.putImageData(oldPixels, 0, 0);
   };   

   /**
    * Sets the mouse cursor style.
    * 
    * @param    {String} style     
    *           Denotes the predefined cursor style.
    */
   Window.prototype.setCursorStyle = function(style)
   {
      this.canvas.style.cursor = style;
   };
   
   /**
    * Sets the stroke style.
    *                              
    * @param    {Number} strokeStyle
    *           The predefined stroke style.
    */
   Window.prototype.setStrokeStyle = function(strokeStyle)
   {
      this.strokeStyle = strokeStyle;
   };
   
   /**
    * Sets the stroke line width.
    *                              
    * @param    {Number} width
    *           The stroke line width.
    */
   Window.prototype.setStrokeWidth = function(width)
   {
      this.strokeWidth = width;    
   };

   /**
    * Sets the line stroke.
    * 
    * @param    {Number} strokeStyle
    *           The predefined stroke style.
    * @param    {Number} width
    *           The stroke line width.
    */
   Window.prototype.setLineStroke = function(strokeStyle, width)
   {                               
      this.setStrokeStyle(strokeStyle);
      this.setStrokeWidth(width);
   };

   /**
    * Draw the given list of operations in the canvas.
    *
    * @param    {byte[]} message
    *           The encoded list of drawing operations.
    */
   Window.prototype.draw = function(message)
   {                               
      var numOps = p2j.socket.readInt32BinaryMessage(message, 5);
                                   
      console.log("START DRAWING CYCLE");
      
      var idx = 9;
      var clips = 0;               
      var trans = 0;
      
      drawLoop:
      for (var i = 0; i < numOps; i++)
      {
         var sz = p2j.socket.readInt32BinaryMessage(message, idx);
                                   
         idx += 4;
         
         var type = message[idx];
         
         var x;
         var y;                    
         var height;
         var width;
         var diameter;
         var raised;
         var text;
         var textLength;           
         var offset;
         var centered;
         var fontId;               
         var fname;
         var path;
         var extra = "";
                                   
         // the following are implemented on the server-side and require no additional
         // implementation here:
         // ops.CLEAR_AREA         
         // ops.CLEAR:
         // ops.FLUSH
         // ops.START_BATCH
         // ops.END_BATCH
               
         switch (type)
         {
            case ops.DRAW_STRING:
               offset = 1;

               textLength = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               text  = p2j.socket.readStringBinaryMessageByLength(message, idx + offset, textLength);
               offset = offset + textLength * 2;

               centered = (p2j.socket.readInt32BinaryMessage(message, idx + offset) == 1);
               offset = offset + 4;

               x = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               y = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               if (centered)
               {
                  // in this case, the "y" is the height in which it needs to be vertically centered
                  y = y / 2;
                  
                  this.ctx.textBaseline = 'middle';
               }
               else
               {
                  this.ctx.textBaseline = 'top';
               }

               this.ctx.fillText(text, x, y);
               break;
            case ops.DRAW_STRING_SCALED:
               offset = 1;

               textLength = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               text  = p2j.socket.readStringBinaryMessageByLength(message, idx + offset, textLength);
               offset = offset + textLength * 2;

               centered = (p2j.socket.readInt32BinaryMessage(message, idx + offset) == 1);
               offset = offset + 4;

               x = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               y = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               var lwidth = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               var lheight = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;
               
               var textWidth = p2j.fonts.getTextWidth(currentFont, text);
               var textHeight = p2j.fonts.getTextHeight(currentFont, text);
               
               var widthScale = lwidth / textWidth;
               var heightScale = lheight / textHeight;
               
               // scale drawing context to computed width
               // this scales the desired structure.x as well so it needs adjusting
               this.ctx.save();
               this.ctx.scale(widthScale, heightScale);
               
               var scaleBackWidth = 1 / widthScale;
               var scaleBackHeight = 1 / heightScale;
               
               if (centered)
               {
                  // in this case, the "y" is the height in which it needs to be vertically centered
                  y = y / 2;

                  this.ctx.textBaseline = 'middle';
               }
               else
               {
                  this.ctx.textBaseline = 'top';
               }

               this.ctx.fillText(text, x * scaleBackWidth, y * scaleBackHeight);
               this.ctx.restore();
               
               break;
            case ops.DRAW_PARAGRAPH:
               offset = 1;

               textLength = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               text  = p2j.socket.readStringBinaryMessageByLength(message, idx + offset, textLength);
               offset = offset + textLength * 2;

               fontId = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               x = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               y = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;
               
               width = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               this.ctx.textBaseline = 'top';

               me.layoutParagraphWorker(text, fontId, x, y, width, true);

               break;
            case ops.DRAW_LINE:
               var x1 =  p2j.socket.readInt32BinaryMessage(message, idx + 1);
               var y1 =  p2j.socket.readInt32BinaryMessage(message, idx + 5);
               var x2 =  p2j.socket.readInt32BinaryMessage(message, idx + 9);
               var y2 =  p2j.socket.readInt32BinaryMessage(message, idx + 13);
               path = this.drawLine(this.ctx, x1, y1, x2, y2, this.rawColor);
               
               strokesManager.applyStrokeToPath(this.ctx,
                                                this.strokeStyle,
                                                this.strokeWidth,
                                                this.rawColor,
                                                path);
               this.ctx.stroke();
               break;
            case ops.DRAW_RECT:
               x      = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y      = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               path   = this.drawRect(this.ctx, x, y, width, height, this.rawColor, false);
               
               strokesManager.applyStrokeToPath(this.ctx,
                                                this.strokeStyle,
                                                this.strokeWidth,
                                                this.rawColor,
                                                path);
               break;
            case ops.DRAW_ROUND_RECT:
               x        = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y        = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width    = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height   = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               diameter = p2j.socket.readInt32BinaryMessage(message, idx + 17);
               this.drawRoundRect(ctx, x, y, width, height, diameter, this.rawColor, false);
               break;
            case ops.DRAW_IMAGE:
               //console.debug("drawing type: " + message[idx]);
               x = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               //console.debug(x + "x" + y + ", " + width + "x" + height);
               var encoding = message[idx + 17];
               //console.debug("encoding type: " + encoding);
               var key = p2j.socket.readInt32BinaryMessage(message, idx + 18);
               console.debug("hash: " + key);
               var pixelsInBytes = width * height * 4;
               var imgData;
               var imgDataOffset;
               if(loadedImages.has(key))
               {
                  imgData = loadedImages.get(key);
                  imgDataOffset = 0;
               }
               else
               {
                  imgData = message;
                  imgDataOffset = idx + 22;
                  loadedImages.set(key, message.subarray(imgDataOffset, imgDataOffset + pixelsInBytes));
               }
               this.drawImage(this.ctx, x, y, width, height, imgData, imgDataOffset);
               break;
            case ops.FILL_RECT:
               x      = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y      = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               this.drawRect(this.ctx, x, y, width, height, this.rawColor, true);
               break;
            case ops.FILL_ROUND_RECT:
               x        = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y        = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width    = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height   = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               diameter = p2j.socket.readInt32BinaryMessage(message, idx + 17);
               this.drawRoundRect(this.ctx, x, y, width, height, diameter, this.rawColor, true);
               break;
            case ops.FILL_POLYGON:
               offset = idx + 1;
               var numPoints = p2j.socket.readInt32BinaryMessage(message, offset);
               
               var xPoints = new Array();
               var yPoints = new Array();
               
               for (var j = 0; j < numPoints; j++)
               {
                  offset += 4;
                  xPoints[j] = p2j.socket.readInt32BinaryMessage(message, offset);
                  offset += 4;
                  yPoints[j] = p2j.socket.readInt32BinaryMessage(message, offset);
               }
               
               this.drawPolygon(this.ctx, xPoints, yPoints, numPoints, this.rawColor, true);
               break;
            case ops.DRAW_3D_RECT:
               x      = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y      = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               raised = (message[idx + 17] === 0);
               this.draw3DRect(this.ctx, x, y, width, height, this.rawColor, false, raised);
               break;
            case ops.FILL_3D_RECT:
               x      = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y      = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               raised = (message[idx + 17] === 0);
               this.draw3DRect(this.ctx, x, y, width, height, this.rawColor, true, raised);
               break;
            case ops.SET_COLOR:
               var r  = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               var g  = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               var b  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               this.setColor([ r, g, b ]);
               break;   
            case ops.SET_LINE_STROKE:
               var strokeStyle = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               var strokeWidth = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               this.setLineStroke(strokeStyle, strokeWidth);
               break;              
            case ops.SET_FONT:
               fontId = p2j.socket.readInt32BinaryMessage(message, idx + 1);

               setFont(this.ctx, fontId);
               break;
            case ops.SET_FONT_STYLE:
               var style = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               
               if (p2j.fonts.setFontStyle(currentFont, style))
               {
                  fname = p2j.fonts.getFontName(currentFont);

                  this.ctx.font = fname;
               }
               break;
            case ops.TRANSLATE_PUSH:                                                   
               x =  p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y =  p2j.socket.readInt32BinaryMessage(message, idx + 5);
               trans++;
               extra = " " + trans.toString() + "; x = " + x.toString() +"; y = " + y.toString();
               this.translate(x, y);  
               break;
            case ops.TRANSLATE_POP:                                                   
               x =  p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y =  p2j.socket.readInt32BinaryMessage(message, idx + 5);
               trans--;
               extra = " " + trans.toString() + "; x = " + x.toString() +"; y = " + y.toString();
               this.untranslate(x, y);
               break;
            case ops.CLIP:
               x      = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y      = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               clips++;
               extra = " " + clips.toString() + 
                       "; x = " + x.toString() +
                       "; y = " + y.toString() +
                       "; width = " + width.toString() + 
                       "; height = " + height.toString();
               this.clip(x, y, width, height);
               break;
            case ops.NO_CLIP:
               clips--;
               extra = " " + clips.toString();
               this.unclip();
               break;
            case ops.SET_LOCATION:
               x =  p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y =  p2j.socket.readInt32BinaryMessage(message, idx + 5);
               this.canvas.style.left = x.toString() + "px";
               this.canvas.style.top  = y.toString() + "px";
               break;
            case ops.RESIZE_WINDOW:
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               this.resize(width, height);
               break;
            case ops.DRAW_HIGHLIGHT:
               break;
            case ops.CLEAR_HIGHLIGHT:
               break;
            case ops.SET_TITLE:
               var title = p2j.socket.readStringBinaryMessage(message, idx + 1);
               this.title = title;
               // TODO: force taskbar to repaint
               break;
            default:
               if (typeof type !== "undefined")
               {
                  me.error("Unknown operation " + type.toString() + "!");
               }
               else
               {
                  me.error("Your session is corrupted buddy!");
                  /*
                  var binMsg  = '';
                  for (var j = 0; j < message.length; j++)
                  {
                     binMsg += String.fromCharCode(message[j]);
                  }
                  var b64msg = window.btoa(binMsg);
                  console.log(b64msg);
                  */
               }
               console.trace();
               break drawLoop;
         }
         
         console.log("PaintPrimitives.%s %s\n", ops[type], extra);
         
         idx += sz;
      }
      
      console.log("END DRAWING CYCLE");
   }; 
   
   /**
    * Checks if the given value is a valid integer that is greater than 0.
    *
    * @param    wid
    *           The value to test.
    *
    * @return   <code>false</code> if the given value is <code>null</code>,
    *           <code>undefined</code>, not an integer or is non-positive.  
    */
   function isValidWindowId(wid)
   {
      return (typeof wid === 'number') && (wid % 1 === 0) && wid > 0;
   };
   
   /**
    * Checks if the given value already exists as a window id.
    *
    * @param    wid
    *           The value to test.
    *
    * @return   <code>false</code> if the given value does not exist in the window list.
    */
   function isExistingWindowId(wid)
   {
      return winlist[wid] instanceof Window;
   };

   /**
    * Obtain the specified window.
    *
    * @param    wid
    *           The window id to obtain.
    *
    * @return   The window instance or <code>null</code> if a window of the given id does not
    *           exist in the window list.
    */
   function getWindow(wid)
   {
      return winlist[wid] instanceof Window ? winlist[wid] : null;
   };

   /**
    * Find the index of the given window in the z-order list.
    *
    * @param    {Number} wid
    *           The window ID to lookup.
    */
   function findZOrderEntry(wid)
   {
      var found = -1;
      
      // find the window in the z-order list
      for (var zidx = 0; 0 < zlist.length; zidx++)
      {
         // we keep the array non-sparse so there should never be a missing element
         if (zlist[zidx].id != wid)
         {
            // the id doesn't match, check the next one
            continue;
         }
         else
         {
            // found it!
            found = zidx;
            
            break;
         }
      }
      
      if (found == -1)
      {
         me.error("Missing window " + wid.toString() + " in z-order list.");
      }
      
      return found;
   };

   /**
    * Adds the given window to the end of the z-order list, setting the z-order value as needed.
    *
    * @param    {Number} wid
    *           The window ID to add.
    * @param    {Window} win
    *           The window instance.
    */
   function addZOrderEntry(wid, win)
   {
      // add to the end of the list
      zlist.push({ "id" : wid, "win" : win });

      // the z-order is based on the array index
      var zidx = zlist.length * 10;
      
      // assign the z-order
      win.canvas.style.zIndex = zidx;
   };
   
   /**
    * Removes the given window from the z-order list and resets the z-order values of all
    * windows that are above the given window.  This compresses the z-index values and
    * eliminates any gaps.
    *
    * @param    {Number} wid
    *           The window ID to remove.
    *
    * @return   {id : Number , win : Window} The removed entry.
    */
   function removeZOrderEntry(wid)
   {
      var found = findZOrderEntry(wid);
      
      if (found == -1)
      {
         // error was already raised
         return;
      }
      
      return removeZOrderEntryWorker(found);
   };
      
   /**
    * Removes the given element from the z-order list and resets the z-order values of all
    * windows that are above the given window.  This compresses the z-index values and
    * eliminates any gaps.
    *
    * @param    {Number} idx
    *           The index of the element to remove.
    *
    * @return   {id : Number , win : Window} The removed entry.
    */
   function removeZOrderEntryWorker(idx)
   {
      // delete the entry and maintain the non-sparse nature of the list
      var removed = zlist.splice(found, 1);
      
      // all subsequent entries are kept in the same order, but we compress the z-index
      // values so there are no gaps
      for (var midx = found; midx < zlist.length; midx++)
      {
         // the z-order is based on the array index
         zlist[midx].win.canvas.style.zIndex = midx * 10;
      }
      
      return removed;
   };
   
   /**
    * Moves the given window to the top of the z-order list and resets the z-order values of all
    * windows that are now moved below the given window.  This compresses the z-index values and
    * eliminates any gaps.
    *
    * @param    {Number} wid
    *           The window ID to move.
    */
   function moveZOrderEntryToTop(wid)
   {
      var last  = zlist.length - 1;
      var found = findZOrderEntry(wid);
      
      if (found != last)
      {
         var entry = removeZOrderEntryWorker(found);
         
         addZOrderEntry(entry.id, entry.win);
      }
   };
   
   /**
    * Directly draw the specified pixel in the given color.
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context on which to draw.
    * @param    {Number} x
    *           X coordinate of the pixel to be drawn.
    * @param    {Number} y
    *           Y coordinate of the pixel to be drawn.
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    */
   Window.prototype.drawPixel = function(ctx, x, y, color)
   {
      this.pixData[0] = color[0];
      this.pixData[1] = color[1];
      this.pixData[2] = color[2];
      
      this.ctx.putImageData(this.pixel, x, y);
   };
   
   /**
    * Draw a line in the given color from (x1, y1) to (x2, y2) inclusive, using direct drawing.
    * <p>
    * This is an implementation of the well known 50+ year old Bresenham algorithm.
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context on which to draw.
    * @param    {Number} x1
    *           X coordinate of the starting pixel to be drawn.
    * @param    {Number} y1
    *           Y coordinate of the starting pixel to be drawn.
    * @param    {Number} x2
    *           X coordinate of the ending pixel to be drawn.
    * @param    {Number} y2
    *           Y coordinate of the ending pixel to be drawn.
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    * 
    * @return   {Array} path
    *           The {x:.,y:.} point per a pixel array that represents the rectangle outerline.
    */
   Window.prototype.drawLine = function(ctx, x1, y1, x2, y2, color)
   {
      x1 = x1 + this.origin.x;
      x2 = x2 + this.origin.x;
      
      y1 = y1 + this.origin.y;
      y2 = y2 + this.origin.y;

      // holds the line pixels to draw
      var path = [];
      var dx = Math.abs(x2 - x1);
      var dy = Math.abs(y2 - y1);
      
      var xIncr = x1 < x2 ? 1 : -1;
      var yIncr = y1 < y2 ? 1 : -1;
      
      dy = -dy;
      
      var err = dx + dy;
      var err2;
      
      var x = x1;
      var y = y1;
      
      var tooMany = 0;
      
      while (true)
      {
         if (strokesManager.isDirectDrawingStrokeStyle(this.strokeStyle))
         {
            this.drawPixel(ctx, x, y, color);
         }
         path.push({x : x, y : y});
         
         tooMany++;
         
         if (x == x2 && y == y2)
         {
            break;
         }
         else
         {
            if (tooMany > 10000)
            {
               console.log("Bresenham is out of control for x1 %d, y1 %d, x2 %d, y2 %d and " +
                           "current x %d, y %d!",
                           x1,
                           y1,
                           x2,
                           y2,
                           x,
                           y);
               
               console.trace();
               break;
            }
         }
            
         err2 = err * 2;
         
         if (err2 >= dy)
         {
            err += dy;
            x   += xIncr; 
         }
         
         if (err2 <= dx)
         {
            err += dx;
            y   += yIncr;
         }
      }
      
      return path;
   };
   
   /**
    * Fill or stroke the current path on the given context, first using the current path as a
    * clipping region.  This ensures that the drawing of the path will not extend outside of the
    * path by a half-pixel (which it would otherwise). The state of the context is saved on
    * entry and restored after the drawing operation, so there will be no residual clipping
    * region when this function returns.
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context on which to draw.
    * @param    {Boolean} fill
    *           <code>true</code> to fill the drawn rectangle with the given color.
    */
   function renderClosedPath(ctx, fill)
   {
      ctx.save();
      ctx.clip();
      
      if (fill)
      {
         ctx.fill();
      }
      else
      {
         ctx.stroke();
      }
      
      ctx.restore();
   };

   /**
    * Draw a rectangle in the given color and dimensions with the line drawing being overwritten
    * using direct drawing to eliminate the negative/unwanted effects of alni-aliasing.
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context on which to draw.
    * @param    {Number} x
    *           X coordinate of the top left corner.
    * @param    {Number} y
    *           Y coordinate of the top left corner.
    * @param    {Number} width
    *           Width of the rectangle in pixels.
    * @param    {Number} height
    *           Height of the rectangle in pixels.
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *           This will be used for the border lines and when in fill mode, also the interior.
    * @param    {Boolean} fill
    *           <code>true</code> to fill the drawn rectangle with the given color.
    * 
    * @return   {Array} path
    *           The {x:.,y:.} point per a pixel array that represents the rectangle outerline.
    */
   Window.prototype.drawRect = function(ctx, x, y, width, height, color, fill)
   {
      // use vector operations for the interior of the rectangle
      if (fill)
      {
         ctx.beginPath();
         ctx.rect(x, y, width, height);
         renderClosedPath(ctx, fill);
      }
      
      // now overdraw the stroked portion to eliminate anit-aliasing
      // to draw rectangle following a clockwise direction
      var path = this.drawLine(ctx, x, y, x + width, y, color);
      Array.prototype.push.apply(path, this.drawLine(ctx, x + width, y + 1, x + width, y + height, color));
      Array.prototype.push.apply(path, this.drawLine(ctx, x + width - 1, y + height, x + 1, y + height, color));
      Array.prototype.push.apply(path, this.drawLine(ctx, x, y + height, x, y + 1, color));// close the path
      
      return path;
   };
   
   /**
    * It draws an image on the canvas.
    * 
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context on which to draw.
    * @param    {Number} x
    *           X coordinate of the top left corner.
    * @param    {Number} y
    *           Y coordinate of the top left corner.
    * @param    {Number} width
    *           Width of the image in pixels.
    * @param    {Number} height
    *           Height of the image in pixels.
    * @param    {Array} imgData
    *           The image data array contains the encoded image started from the given offset.
    * @param    {Number} imgDataOffset
    *           The offset of the encoded image data.
    */
   Window.prototype.drawImage = function(ctx, x, y, width, height, imgData, imgDataOffset)
   {
      var img = ctx.createImageData(width, height);
      var data = img.data;
      var pixelsInBytes = width * height * 4;

      for (var i = 0, j = imgDataOffset; i < pixelsInBytes; i += 4, j += 4)
      {
         data[i]     = imgData[j];
         data[i + 1] = imgData[j + 1];
         data[i + 2] = imgData[j + 2];
         data[i + 3] = imgData[j + 3];
     }
     ctx.putImageData(img, this.origin.x + x, this.origin.y + y);
   };
   
   /**
    * Convert a standard fillStyle or strokeStyle color into an array of [ R, G, B ] values.
    *
    * @param    {String} spec
    *           Input string which must be in the form "#RRGGBB" where RR, GG and BB are 2-digit
    *           hexadecimal numbers.
    *
    * @return   The parsed raw color array of integers in [ R, G, B ] form.
    */
   function parseColor(spec)
   {
      if (typeof spec !== "string")
      {
         me.error("parseColor() spec must be a string");
         return null;                                 
      }
      
      if (spec.length !== 7)
      {
         me.error("parseColor() spec must have length of 7");
         return null;
      }
      
      if (spec.charAt(0) !== "#")
      {
         me.error("parseColor() spec must start with #");
         return null;
      }
      
      var r = parseInt(spec.substr(1, 2), 16);
      var g = parseInt(spec.substr(3, 2), 16);
      var b = parseInt(spec.substr(5, 2), 16);
   
      if (r === NaN || g === NaN || b === NaN)
      {
         return null;
      }
      
      return [ r, g, b ];
   };
   
   /**
    * Test if the given argument is an array of 3 integer values.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {Boolean} <code>true</code> if the argument is in [ R, G, B ] form.
    */
   function validateColor(color)
   {
      if (!(color instanceof Array))
      {
         return false;
      }
      
      if (color.length !== 3)
      {
         return false;
      }
      
      if (typeof color[0] !== "number" ||
          typeof color[1] !== "number" ||
          typeof color[2] !== "number")
      {
         return false;
      }
      
      if (color[0] < 0 || color[0] > 255 ||
          color[1] < 0 || color[1] > 255 ||
          color[2] < 0 || color[2] > 255)
      {
         return false;
      }
      
      return true;
   };
   
   /**
    * Convert the given array of 3 integer values into an HTML color string.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {String} A valid HTML color string.
    */
   function createColorString(color)
   {
      if (!validateColor(color))
      {
         return null;
      }
                                         
      return "rgb(" + color[0].toString() + ", "
                    + color[1].toString() + ", "
                    + color[2].toString() + ")";
   };
   
   /**
    * Convert the given array of RGB values into a lighter color in a manner that is compatible 
    * with Java2D.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {Number[]} A lighter shade of the same color.
    */
   function lightenColor(color)
   {
      if (!validateColor(color))
      {
         return null;
      }
      
      // handle black directly, the result is NOT very much different but this is how it is
      // done in Java
      if (color[0] == 0 && color[1] == 0 && color[2] == 0)
         return [ MIN_COLOR_VALUE, MIN_COLOR_VALUE, MIN_COLOR_VALUE ];
      
      // force inputs to a minimim color value if between 0 and that minimum color value
      // (yes, this is how Java does it)
      var rIn = (color[0] > 0 && color[0] < MIN_COLOR_VALUE) ? MIN_COLOR_VALUE : color[0];
      var gIn = (color[1] > 0 && color[1] < MIN_COLOR_VALUE) ? MIN_COLOR_VALUE : color[1];
      var bIn = (color[2] > 0 && color[2] < MIN_COLOR_VALUE) ? MIN_COLOR_VALUE : color[2];
      
      // scale it
      var r = Math.min(Math.floor(rIn / SCALE_FACTOR), 255);
      var g = Math.min(Math.floor(gIn / SCALE_FACTOR), 255);
      var b = Math.min(Math.floor(bIn / SCALE_FACTOR), 255);
      
      return [ r, g, b ];
   };
   
   /**
    * Convert the given array of RGB values into a darker color in a manner that is compatible 
    * with Java2D.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {Number[]} A darker shade of the same color.
    */
   function darkenColor(color)
   {
      if (!validateColor(color))
      {
         return null;
      }
      
      var r = Math.max(0, Math.floor(color[0] * SCALE_FACTOR));
      var g = Math.max(0, Math.floor(color[1] * SCALE_FACTOR));
      var b = Math.max(0, Math.floor(color[2] * SCALE_FACTOR));
      
      return [ r, g, b ];
   };
         
   /**
    * Draw a rounded rectangle with the given context, dimensions and fill.
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context.
    * @param    {Number} x
    *           X coordinate of the top left of the rectangle (if a square corner was placed
    *           there).
    * @param    {Number} y
    *           Y coordinate of the top left of the rectangle (if a square corner was placed
    *           there).
    * @param    {Number} width
    *           Width of the rectangle.
    * @param    {Number} height
    *           Height of the rectangle.
    * @param    {Number} diameter
    *           Diameter of the arc that defines the rounded corner.
    * @param    {Number[]} color  
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    * @param    {Boolean} fill
    *           <code>true</code> to fill the rectangle with the current color, otherwise just
    *           stroke the rectangle.
    */
   Window.prototype.drawRoundRect = function(ctx, x, y, width, height, diameter, color, fill)
   {
      var radius = diameter / 2;
      
      x = x + this.origin.x;
      y = y + this.origin.y;
      
      // arc control points
      var northWestX = x;
      var northWestY = y;
      var northEastX = x + width;
      var northEastY = y;
      var southWestX = x;
      var southWestY = y + height;
      var southEastX = x + width;
      var southEastY = y + height;
      
      // line end points                                       
      var topLeftX     = northWestX + radius;
      var topLeftY     = northWestY;
      var topRightX    = northEastX - radius;
      var topRightY    = northEastY;
      var rightUpX     = northEastX;
      var rightUpY     = northEastY + radius;
      var rightDownX   = southEastX;
      var rightDownY   = southEastY - radius;
      var bottomRightX = southEastX - radius;
      var bottomRightY = southEastY;
      var bottomLeftX  = southWestX + radius;
      var bottomLeftY  = southWestY;
      var leftDownX    = southWestX;
      var leftDownY    = southWestY - radius;
      var leftUpX      = northWestX;
      var leftUpY      = northWestY + radius;
      
      ctx.beginPath();
      ctx.moveTo(topRightX, topRightY);
      ctx.quadraticCurveTo(northEastX, northEastY, rightUpX, rightUpY);
      ctx.lineTo(rightDownX, rightDownY);
      ctx.quadraticCurveTo(southEastX, southEastY, bottomRightX, bottomRightY);
      ctx.lineTo(bottomLeftX, bottomLeftY);
      ctx.quadraticCurveTo(southWestX, southWestY, leftDownX, leftDownY);
      ctx.lineTo(leftUpX, leftUpY);
      ctx.quadraticCurveTo(northWestX, northWestY, topLeftX, topLeftY);
      ctx.closePath();
      
      renderClosedPath(ctx, fill);
      
      // overdraw the anti-aliased line segments              
      this.drawLine(ctx, topLeftX, topLeftY, topRightX, topRightY, color);
      this.drawLine(ctx, rightUpX, rightUpY, rightDownX, rightDownY, color);
      this.drawLine(ctx, bottomLeftX, bottomLeftY, bottomRightX, bottomRightY, color);
      this.drawLine(ctx, leftUpX, leftUpY, leftDownX, leftDownY, color);
   };
   
   /**
    * Draw a slightly shaded "3D" rectangle with the given context, dimensions and fill. In
    * raised mode the rectangle is drawn with a lighter color used for the top and left borders
    * and any fill that is needed, while the bottom and right borders are a darker version of
    * the given color.  If not raised, a darker shade of the given color is used for the top
    * and left borders and any requested fill, while a lighter shade is used for the bottom and
    * right borders.
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context.
    * @param    {Number} x
    *           X coordinate of the top left of the rectangle.
    * @param    {Number} y
    *           Y coordinate of the top left of the rectangle.
    * @param    {Number} width
    *           Width of the rectangle in pixels.
    * @param    {Number} height
    *           Height of the rectangle in pixels.
    * @param    {Number[]} color  
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    * @param    {Boolean} fill
    *           <code>true</code> to fill the rectangle with the current color, otherwise just
    *           stroke the rectangle.
    * @param    {Boolean} raise
    *           <code>true</code> to draw in raised mode.
    */
   Window.prototype.draw3DRect = function(ctx, x, y, width, height, color, fill, raised)
   {
      // save off our colors
      var oriFillColor   = ctx.fillStyle;
      var oriStrokeColor = ctx.strokeStyle;
      
      var darker  = darkenColor(color);
      var lighter = lightenColor(color);
      
      // save off our state so that we can clear our clipping region on exit
      ctx.save();
      
      // the Java implementation draws 1 pixel larger in each dimension
      width  = width + 1;
      height = height + 1;
      
      // setup our clipping region, this is needed because the line width (1 pixel wide) is
      // drawn half on one side of hte path and half on the other side; this makes paths a
      // half pixel bigger than the caller would expect; by defining the same path as a
      // clipping region, the half pixel width that is outside the path is never output
      ctx.beginPath();
      ctx.rect(x, y, width, height);
      ctx.clip();
      
      var current = raised ? lighter : darker;
      
      if (fill)
         ctx.fillStyle = createColorString(current);
         
      this.drawRect(ctx, x, y, width, height, current, fill);
      
      // draw the bottom and right sides in the contrasting color
      current = raised ? darker : lighter;
      this.drawLine(ctx, x + 1, y + height, x + width, y + height, current);
      this.drawLine(ctx, x + width, y, x + width, y + height - 1, current);
      
      // clear the clipping region
      ctx.restore();
      
      // restore our colors
      ctx.fillStyle   = oriFillColor;
      ctx.strokeStyle = oriStrokeColor;
   };

   /**
    * Draw a closed polygon with the given context, vertices and fill.
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context.
    * @param    {Number} xPoints
    *           X coordinates of the points.
    * @param    {Number} yPoints
    *           Y coordinates of the points.
    * @param    {Number} num
    *           Number of points to draw.
    * @param    {Number[]} color  
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    * @param    {Boolean} fill
    *           <code>true</code> to fill the rectangle with the current color, otherwise just
    *           stroke the rectangle.
    */
   Window.prototype.drawPolygon = function(ctx, xPoints, yPoints, num, color, fill)
   {
      var i;
   
      // use vector operations for the interior of the poygon
      if (fill)
      {
         ctx.beginPath();
         ctx.moveTo(xPoints[0] + this.origin.x, yPoints[0] + this.origin.y);
         for (i = 1; i < num; i++)
         {
            ctx.lineTo(xPoints[i] + this.origin.x, yPoints[i] + this.origin.y);
         }
         renderClosedPath(ctx, fill);
      }
      
      // now overdraw the stroked portion
      for (i = 0; i < (num - 1); i++)
      {
         this.drawLine(ctx, xPoints[i], yPoints[i], xPoints[i + 1], yPoints[i + 1], color);
      }
   };
   
   /**
    * Set the font in the current context.  If the current font does not get changed, then this
    * will be a no-op.
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context on which to draw.
    * @param    {Number}  fontId
    *           The font identifier.
    */
   function setFont(ctx, fontId)
   {
      if (fontId != currentFont)
      {
         var fname = p2j.fonts.getFontName(fontId);
      
         ctx.font = fname;
      
         currentFont = fontId;
      }
   };

   /**
    * Find the next split position, so that the sub-text fits the given width.
    *
    * @param    {Number} font
    *           The font used to measure the text.
    * @param    {String} txt
    *           The paragraph text.
    * @param    {Number} start
    *           The index where to start splitting.
    * @param    {Number} width
    *           The width in which the sub-text must fit.
    *
    * @return   {Number} The end split position, exclusive.
    */
   function findNextPos(font, txt, start, width)
   {
      var end = start;
      do
      {
         var nextWord = regexIndexOf(txt, /\s/g, end);

         if (nextWord == -1)
         {
            // no more words, use the entire text from "start"
            nextWord = txt.length;
         }

         var candidate = txt.substring(start, nextWord);
         var cwidth = p2j.fonts.getTextWidth(font, candidate);

         if (cwidth > width)
         {
            // two cases
            if (end == start)
            {
               // if it's the start of the line, find where to split the word
               end = splitWord(font, txt, start, width);
            }
            else
            {
               // not the start of the line, break on the previous word
            }

            break;
         }

         // the line is not yet filled, continue
         end = nextWord + 1;
      }
      while (end < txt.length);

      return end;
   };

   /**
    * Split the word starting on the given index so that it fits the specified width.
    *
    * @param    {Number} font
    *           The font used to measure the text.
    * @param    {String} txt
    *           The paragraph text.
    * @param    {Number} start
    *           The index where to start splitting.
    * @param    {Number} width
    *           The width in which the sub-text must fit.
    *
    * @return   {Number} The end split position, exclusive.
    */
   function splitWord(font, txt, start, width)
   {
      var end = start + 1;
      while (end < txt.length)
      {
         var candidate = txt.substring(start, end);
         var cwidth = p2j.fonts.getTextWidth(font, candidate);
         
         if (cwidth > width)
         {
            break;
         }

         end = end + 1;
      }

      return end;
   };

   /**
    * Find the next occurrence of the character given by the specified regex, starting from the
    * specified index.
    *
    * @param    {String} txt
    *           The text.
    * @param    {regex} regex
    *           The regular expression used for search.
    * @param    {Number} start
    *           The index where to start searching.
    *
    * @return   {Number} The next occurrent of the given regex, or -1 if not found.
    */
   function regexIndexOf(txt, regex, start)
   {
      var indexOf = txt.substring(start).search(regex);
      return (indexOf >= 0) ? (indexOf + start) : indexOf;
   };
   
   /**
    * Create a new top-level window (and the canvas that backs it).
    *
    * @param    wid
    *           The window id by which this window can be uniquely identified.
    *
    * @return   The created window instance or <code>null</code> on any error.
    */
   me.createWindow = function(wid)
   {
      if (!isValidWindowId(wid))
      {
         me.error("Window ids must be non-positive integers.");
         return null;
      }
      
      if (isExistingWindowId(wid))
      {
         me.error("Cannot create window with id " + wid.toString() + " because it already exists.");
         return null;                       
      }
      
      var newwin = new Window(wid);
      
      // store in our window list
      winlist[wid] = newwin;
      
      // add to the z-order list
      addZOrderEntry(wid, newwin);
      
      return newwin;
   };
   
   /**
    * Create a new child window (and the canvas that backs it).
    *
    * @param    wid
    *           The window id by which this window can be uniquely identified.
    * @param    owner
    *           The id of the owning window.
    * @param    title
    *           The title text for the new window.
    *
    * @return   The created window instance or <code>null</code> on any error.
    */
   me.createChildWindow = function(wid, owner, title)
   {
      if (!isValidWindowId(owner))
      {
         me.error("Owner window ids must be non-positive integers.");
         return null;
      }
      
      if (!isExistingWindowId(owner))
      {
         me.error("Owner window with id " + wid.toString() + " must already exist.");
         return null;
      }
      
      var newwin = me.createWindow(wid);
      
      if (newwin === null)
         return null;
      
      newwin.owner = owner;
      newwin.title = title || "";
      
      var parent = getWindow(owner);
      
      // add to the owned window list
      parent.children.push(wid);
   };
   
   /**
    * Destroy the given window (and all resources that back it).
    *
    * @param    wid
    *           The window id by which this window can be uniquely identified.
    * @param    images
    *           The cached images to remove.
    */
   me.destroyWindow = function(wid, images)
   {
      if (!isValidWindowId(wid))
      {
         me.error("Cannot destroy window for id that is not a non-positive integer.");
         return;
      }
      
      if (!isExistingWindowId(wid))
      {
         me.error("Cannot destroy window with id " + wid.toString() + " because it doesn't exist.");
         return;
      }
      
      var win = getWindow(wid);
      
      win.cleanup();
      
      for (var image in images)
      {
         loadedImages.delete(image);
      }
      
      // remove from our window list
      delete winlist[wid];
      
      // remove the window from the z-order
      removeZOrderEntry(wid);
   };
   
   /**
    * Change the visibility of the given window.
    *
    * @param    {int} wid
    *           The window id by which this window can be uniquely identified.
    * @param    {boolean} visible
    *           Flag that determines the new visibility state.
    */
   me.setWindowVisible = function(wid, visible)
   {
      if (!isValidWindowId(wid))
      {
         me.error("Cannot change visibility for window for id that is invalid.");
         return;
      }
      
      if (!isExistingWindowId(wid))
      {
         me.error("Cannot change visibility for non-existent window with id " + wid.toString());
         return;
      }
      
      var win = getWindow(wid);
      
      win.setVisible(visible);
   };
   
   /**
    * Resets the z-order values of all windows to match the given order.
    *
    * @param    {Number[]} wids
    *           The list of window IDs in the new z-order.
    */
   me.restackZOrderEntries = function(wids)
   {
      if (zlist.length != wids.length)
      {
         me.error("Restacking list must have the same number of elements as the z-order list!");
         return;
      }
      
      // drop the old list
      zlist = new Array();
      
      // re-add each element in order
      for (var i = 0; i < wids.length; i++)
      {
         // wids cannot be a sparse array
         if (!(typeof wids[i] === 'number'))
         {
            me.error("Element " + i.toString() + " in restacking list is invalid!");
            continue;
         }
         
         var win = getWindow(wids[i]);
         
         // each element must be a valid window
         if (win === null)
         {
            me.error("Invalid window id " + wids[i].toString() + " in restacking list.");
            continue;
         }
         
         addZOrderEntry(wids[i], win);
      }
   };
   
   /**
    * Display an error message.
    * 
    * @param    {string} text
    *           The error text to display.
    */
   me.error = function(text)
   {
      console.error(text);
      p2j.sound.beep();
   };

   /**
    * Initialize module.
    *  
    * @param {object} cfg configuration.
    */
   me.init = function(cfg)
   {
   };

   /**
    * Set cursor position.
    */
   me.setCursorPosition = function()
   {
      me.error("setCursorPosition() should not be used in GUI!");
   };

   /**
    * Set cursor visibility
    */
   me.setCursorStatus = function()
   {
      me.error("setCursorStatus() should not be used in GUI!");
   };

   /**
    * Set the mouse cursor to the given predifined cursor style. The cursor style id is mapped
    * to the predefined cursors styles: auto | default | none | context-menu | help | pointer
    *  | progress | wait | cell | crosshair | text | vertical-text | alias | copy | move | no-drop
    *  | not-allowed | e-resize | n-resize | ne-resize | nw-resize | s-resize | se-resize
    *  | sw-resize | w-resize | ew-resize | ns-resize | nesw-resize | nwse-resize | col-resize
    *  | row-resize | all-scroll | zoom-in | zoom-out | grab | grabbing 
    * 
    * @param    {Number} cursorStyleId
    *           Denotes the cursor style id.
    * @param    {Number} wid
    *           Denotes the active window id.
    */
   me.setCursorStyle = function(cursorStyleId, wid)
   {
      var style;
      switch (cursorStyleId)
      {
         case 0:
            style = "default";
            break;
         case 1:
            style = "text";
            break;
         case 2:
            style = "n-resize"
            break;
         case 3:
            style = "ne-resize"
            break;
         case 4:
            style = "e-resize"
            break;
         case 5:
            style = "se-resize"
            break;
         case 6:
            style = "s-resize"
            break;
         case 7:
            style = "sw-resize"
            break;
         case 8:
            style = "w-resize"
            break;
         case 9:
            style = "nw-resize"
            break;
         default:
            style = "default";
      }

      var win = getWindow(wid);

      win.setCursorStyle(style);
   };

   /**
    * Clear the screen.
    */
   me.clear = function()
   {
      me.error("clear() should not be used in GUI!");
   };

   /**
    * Draw the given list of operations in the specified window. This is a completely different
    * implementation than for ChUI, but we are re-using the message type.
    *
    * @param    {byte[]} message
    *           The encoded list of drawing operations.
    */
   me.drawRectangles = function(message)
   {
      var wid = p2j.socket.readInt32BinaryMessage(message, 1);
      
      if (!isValidWindowId(wid))
      {
         me.error("Cannot draw window for id that is invalid.");
         return;
      }                                    
      
      if (!isExistingWindowId(wid))
      {
         me.error("Cannot draw non-existent window with id " + wid.toString());
         return;
      }
      
      var win = getWindow(wid);
      
      win.draw(message);
   };

   /**
    * The method performs a layout operation on the supplied text and returns the resulting 
    * paragraph height while maintaining the supplied maximum width.
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context on which to draw.
    * @param    {String} txt
    *           The paragraph text.
    * @param    {Number} font
    *           The font ID with which the layout/drawing will be performed.
    * @param    {Number} x
    *           The drawing X position, ignored if drawing is not required.
    * @param    {Number} y
    *           The drawing Y position, ignored if drawing is not required.
    * @param    {Number} width
    *           The maximum width used to layout the paragraph.
    * @param    {boolean} draw
    *           When <code>true</code>, this flag indicates that the text should be drawn on the
    *           current context.
    * 
    * @return   {Number} The paragraph height.
    */
   me.layoutParagraphWorker = function(ctx, txt, font, x, y, width, draw)
   {
      var newLine = true;
      var i = 0;
      
      if (draw)
      {
         // unless we are drawing, there is no need in changing the context font
         ctx.save();
      
         setFont(ctx, font);
      }
      
      var height = 0;
      var fheight = p2j.fonts.getFontHeight(font);
      
      while (i < txt.length)
      {
         var nextPos = findNextPos(txt, i, width);

         var line = txt.substring(i, nextPos);

         if (draw)
         {
            ctx.fillText(line, x, y);
         }

         y = y + fheight;
         i = nextPos;
         
         height = height + fheight;
      }
      
      if (draw)
      {
         ctx.restore();
      }
      
      return height;
   }

   return me;
}());
